# -*- coding: utf-8 -*-
import os


def fake_urlretrieve(url, target):
    open(target, 'a').close()


class FakeZipFile(object):

    def __init__(self, in_path, mode='r'):
        self.in_path = in_path

    def extractall(self, target_dir):
        out_path = os.path.join(target_dir, self.in_path.replace('.zip', ''))
        open(out_path, 'a').close()



# -*- coding: utf-8 -*-
import click
import logging

from .run import run_sambamba

log = logging.getLogger(__name__)


@click.command()
@click.option('-r', '--regions', type=click.Path(exists=True), required=True,
              help='Path to a bed file with exon coordinates')
@click.option('-t', '--cov_treshold', multiple=True, type=int,
              help=("multiple thresholds can be provided,"
                    "for each one an extra column will be added,"
                    "the percentage of bases in the region"
                    "where coverage is more than this value"))
@click.option('-o', '--outfile', type=click.Path(exists=False),
              help='Specify path to a file where results should be stored.')
@click.argument('bam_file', type=click.Path(exists=True))
@click.pass_context
def sambamba(context, bam_file, regions, cov_treshold, outfile):
    """Run Sambamba from chanjo."""
    log.info("Running chanjo sambamba")
    try:
        run_sambamba(bam_file, regions, outfile, cov_treshold)
    except Exception as error:
        log.exception('something went really wrong :_(')
        context.abort()



# -*- coding: utf-8 -*-
"""
Parse the sambamba "depth region" output.
"""
from chanjo.compat import iteritems
from chanjo.exc import BedFormattingError


def depth_output(handle):
    """Parse the output.

    Args:
        handle (iterable): Chanjo-formatted BED lines

    Yields:
        dict: parsed sambamba output row

    Raises:
        BedFormattingError: if the BED file doesn't contain enough columns
    """
    lines = (line.strip() for line in handle)
    rows = (line.split('\t') for line in lines)
    # expect only a single header row
    header_row = next(rows)
    if len(header_row) < 6:
        raise BedFormattingError('make sure fields are tab-separated')
    header_data = expand_header(header_row)
    # parse rows
    for row in rows:
        yield expand_row(header_data, row)


def expand_header(row):
    """Parse the header information.

    Args:
        List[str]: sambamba BED header row

    Returns:
        dict: name/index combos for fields
    """
    # figure out where the sambamba output begins
    sambamba_start = row.index('readCount')
    sambamba_end = row.index('sampleName')
    coverage_columns = row[sambamba_start + 2:sambamba_end]
    thresholds = {int(column.replace('percentage', '')): row.index(column)
                  for column in coverage_columns}
    keys = {
        'readCount': sambamba_start,
        'meanCoverage': sambamba_start + 1,
        'thresholds': thresholds,
        'sampleName': sambamba_end,
        'extraFields': slice(3, sambamba_start)
    }
    return keys


def expand_row(header, row):
    """Parse information in row to dict.

    Args:
        header (dict): key/index header dict
        row (List[str]): sambamba BED row

    Returns:
        dict: parsed sambamba output row
    """
    thresholds = {threshold: float(row[key])
                  for threshold, key in iteritems(header['thresholds'])}
    data = {
        'chrom': row[0],
        'chromStart': int(row[1]),
        'chromEnd': int(row[2]),
        'sampleName': row[header['sampleName']],
        'readCount': int(row[header['readCount']]),
        'meanCoverage': float(row[header['meanCoverage']]),
        'thresholds': thresholds,
        'extraFields': row[header['extraFields']]
    }
    return data



import subprocess
import logging

from subprocess import CalledProcessError

log = logging.getLogger(__name__)


def run_sambamba(bam_file, region_file, outfile=None, cov_treshold=()):
    """Run sambamba from chanjo."""
    sambamba_call = [
        "sambamba",
        "depth",
        "region",
        "--regions",
        region_file,
        bam_file
    ]

    if outfile:
        sambamba_call += ['-o', outfile]

    for coverage_treshold in cov_treshold:
        sambamba_call += ['-T', str(coverage_treshold)]

    log.info("Running sambamba with call: %s", ' '.join(sambamba_call))
    try:
        subprocess.check_call(sambamba_call)  # stderr=log_stream
    except OSError as error:
        log.critical("sambamba seems to not exist on your system.")
        raise error
    except CalledProcessError as error:  # pragma: no cover
        log.critical("Something went wrong when running sambamba. "
                     "Please see sambamba error output.")
        raise error

    log.debug("sambamba ran successfully")



# -*- coding: utf-8 -*-
import pytest

from chanjo.load.parse import bed
from chanjo.exc import BedFormattingError


def test_expand_row(bed_row):
    data = bed.expand_row(bed_row)
    assert data['name'] == bed_row[3]

    # test with malformatted values
    # GIVEN position is not int
    bed_row[1] = 'paulthomas'
    with pytest.raises(BedFormattingError):
        bed.expand_row(bed_row)

    bed_row[1] = '100'
    bed_row[4] = '14.3'  # score must be int
    with pytest.raises(BedFormattingError):
        bed.expand_row(bed_row)


def test_expand_row_space_separated():
    with pytest.raises(BedFormattingError):
        bed.expand_row(['22 32588888 32589173 22-32588888-32589173 0 -'])


def test_chanjo(bed_lines):
    exons = list(bed.chanjo(bed_lines))
    assert len(exons) == 19
    exon = exons[0]
    assert exon['chrom'] == '1'
    assert exon['name'] == '1-11-18'


def test_extra_fields():
    # GIVEN no extra columns
    extra_cols = []
    data = bed.extra_fields(extra_cols)
    assert data == []


def test_list_get():
    # GIVEN a regular list with three items
    reg_list = [1, 2, 3]
    # WHEN accsesing existing 2nd item
    item = bed.list_get(reg_list, 1)
    # THEN it should return the list item
    assert item == reg_list[1]

    # WHEN accessing non-existing 4th item
    item = bed.list_get(reg_list, 3)
    # THEN it should return None
    assert item is None

    # WHEN accessing existing item with default
    item = bed.list_get(reg_list, 2, default=10)
    # THEN it should return the item
    assert item == reg_list[2]

    # WHEN accessing non-existing item with default
    default = 10
    item = bed.list_get(reg_list, 4, default=default)
    # THEN it should return the default
    assert item == default

    # GIVEN empty list
    empty_list = []
    # WHEN accessing any item
    item = bed.list_get(empty_list, 0)
    # THEN it should always be None
    assert item is None



# -*- coding: utf-8 -*-
from datetime import datetime

import pytest
from sqlalchemy.orm.exc import FlushError

from chanjo.store.api import ChanjoDB
from chanjo.store.models import Sample


def test_dialect(chanjo_db):
    assert chanjo_db.dialect == 'sqlite'
    assert hasattr(chanjo_db, 'query')


def test_no_dialect():
    # GIVEN not explicity specifying SQLite dialect
    # WHEN setting up the API
    chanjo_db = ChanjoDB(':memory:')
    # THEN chanjo should guess it
    assert chanjo_db.dialect == 'sqlite'


def test_save(chanjo_db):
    # GIVEN a new sample
    sample_id = 'ADM12'
    new_sample = Sample(id=sample_id, group_id='ADMG1', source='alignment.bam')
    # WHEN added and saved to the database
    chanjo_db.add(new_sample)
    chanjo_db.save()
    # THEN is should exist in the database
    assert new_sample.id == sample_id
    assert isinstance(new_sample.created_at, datetime)
    assert Sample.query.get(sample_id) == new_sample

    # GIVEN sample already exists
    conflict_sample = Sample(id=sample_id, group_id='ADMG2')
    # WHEN saving it again with same id
    # THEN error is raised _after_ rollback
    with pytest.raises(FlushError):
        chanjo_db.add(conflict_sample)
        chanjo_db.save()

    new_sampleid = 'ADM13'
    chanjo_db.add(Sample(id=new_sampleid))
    chanjo_db.save()
    assert Sample.query.get(new_sampleid)


def test_add_many(chanjo_db):
    # GIVEN multiple new samples
    new_samples = [Sample(id='ADM12'), Sample(id='ADM13')]
    # WHEN added to the session
    chanjo_db.add(new_samples)
    chanjo_db.save()
    # THEN all samples should be added
    assert Sample.query.all() == new_samples

    # GIVEN an unknown type to add, e.g. dict
    data = {'sample': Sample(id='ADM13')}
    # WHEN adding to session
    # THEN it should raise ValueError
    with pytest.raises(ValueError):
        chanjo_db.add(data)



# -*- coding: utf-8 -*-
from chanjo.cli import root


def test_sex(cli_runner, bam_path):
    result = cli_runner.invoke(root, ['sex', bam_path])
    assert result.exit_code == 0
    assert 'female' in result.output


def test_sex_with_wrong_file(cli_runner, bam_path):
    bai_path = "{}.bai".format(bam_path)
    result = cli_runner.invoke(root, ['sex', bai_path])
    assert result.exit_code != 0



# -*- coding: utf-8 -*-
import pytest

from chanjo.load.parse import sambamba
from chanjo.exc import BedFormattingError


def test_depth_output():
    exon_lines = ['# chrom chromStart chromEnd\n', '1 10 100\n']
    with pytest.raises(BedFormattingError):
        list(sambamba.depth_output(exon_lines))



# -*- coding: utf-8 -*-


def test_sambamba(invoke_cli, bam_path, bed_path):
    # GIVEN a small BAM file and corresponding BED file
    # WHEN running sambamba
    result = invoke_cli(['sambamba', '-r', bed_path, bam_path])
    # THEN command should complete succesfully
    assert result.exit_code == 0


def test_sambamba_with_bai(invoke_cli, bam_path, bed_path):
    # GIVEN a BAI-file instead of BAM
    bai_path = "{}.bai".format(bam_path)
    # WHEN running sambamba
    result = invoke_cli(['sambamba', '-r', bed_path, bai_path])
    # THEN command should exit with error
    assert result.exit_code != 0



# -*- coding: utf-8 -*-

from chanjo.cli import root
from chanjo.store.models import Sample, TranscriptStat


def test_setup(cli_runner, tmpdir):
    # GIVEN an empty directory
    assert tmpdir.listdir() == []
    # WHEN setting up a new database
    db_name = 'coverage.sqlite'
    db_path = tmpdir.join(db_name)
    cli_runner.invoke(root, ['--database', str(db_path), 'db', 'setup'])
    # THEN the file is created
    assert tmpdir.listdir() == [db_path]


def test_setup_reset(cli_runner, popexist_db):
    # GIVEN an existing database with a single sample
    assert Sample.query.count() == 1
    # WHEN resetting it
    cli_runner.invoke(root, ['--database', popexist_db.uri, 'db', 'setup',
                             '--reset'])
    # THEN the database should be empty
    assert Sample.query.count() == 0


def test_remove(cli_runner, popexist_db):
    # GIVEN an existing database with one sample
    sample_id = 'sample'
    assert Sample.query.get(sample_id)
    assert TranscriptStat.query.filter_by(sample_id=sample_id).count() > 0
    # WHEN removing the sample from the CLI
    cli_runner.invoke(root, ['--database', popexist_db.uri, 'db', 'remove',
                             sample_id])
    # THEN the sample should be deleted along with annotations
    assert Sample.query.get(sample_id) is None
    assert TranscriptStat.query.filter_by(sample_id=sample_id).count() == 0

    # WHEN removing a sample with non-existing id
    result = cli_runner.invoke(root, ['--database', popexist_db.uri, 'db',
                                      'remove', 'no-sample-id'])
    # THEN context is aborted
    assert result.exit_code == 1



# -*- coding: utf-8 -*-
import codecs
from functools import partial
import os

from click.testing import CliRunner
import pytest

from chanjo.cli import root
from chanjo.store.api import ChanjoDB
from chanjo.load.parse import bed, sambamba
from chanjo.load.sambamba import load_transcripts
from chanjo.load.link import link_elements


@pytest.yield_fixture
def reset_path():
    """Reset PATH environment variable temporarily."""
    path_env = os.environ['PATH']
    os.environ['PATH'] = ''
    yield path_env
    os.environ['PATH'] = path_env


@pytest.yield_fixture(scope='function')
def chanjo_db():
    _chanjo_db = ChanjoDB('sqlite://')
    _chanjo_db.set_up()
    yield _chanjo_db
    _chanjo_db.tear_down()


@pytest.yield_fixture(scope='function')
def existing_db(tmpdir):
    db_path = tmpdir.join('coverage.sqlite3')
    chanjo_db = ChanjoDB(str(db_path))
    chanjo_db.set_up()
    yield chanjo_db
    chanjo_db.tear_down()


@pytest.yield_fixture(scope='function')
def popexist_db(existing_db, exon_lines):
    result = link_elements(exon_lines)
    existing_db.add(result.models)
    result = load_transcripts(exon_lines, sample_id='sample', group_id='group')
    existing_db.add(result.sample)
    existing_db.add(result.models)
    existing_db.save()
    yield existing_db


@pytest.yield_fixture(scope='function')
def populated_db(chanjo_db, exon_lines):
    result = link_elements(exon_lines)
    chanjo_db.add(result.models)
    result = load_transcripts(exon_lines, sample_id='sample', group_id='group')
    chanjo_db.add(result.sample)
    chanjo_db.add(result.models)
    chanjo_db.save()
    yield chanjo_db


@pytest.fixture
def exon_lines(sambamba_path):
    with codecs.open(sambamba_path, 'r', encoding='utf-8') as stream:
        _exon_lines = [line for line in stream]
    return _exon_lines


@pytest.fixture
def bed_lines(bed_path):
    with codecs.open(bed_path, 'r', encoding='utf-8') as stream:
        _bed_lines = [line for line in stream]
    return _bed_lines


@pytest.fixture
def bed_row():
    row = ['22', '32588888', '32589173', '22-32588888-32589173', '0', '-',
           'CCDS54521.1', 'RFPL2']
    return row


@pytest.fixture
def bed_exons(bed_lines):
    _exons = bed.chanjo(bed_lines)
    return _exons


@pytest.fixture
def sambamba_exons(exon_lines):
    _exons = sambamba.depth_output(exon_lines)
    return _exons


@pytest.fixture
def cli_runner():
    runner = CliRunner()
    return runner


@pytest.fixture
def invoke_cli(cli_runner):
    return partial(cli_runner.invoke, root)


@pytest.fixture
def bam_path():
    _bam_path = 'tests/fixtures/ccds-mini.sorted.bam'
    return _bam_path


@pytest.fixture
def bed_path():
    _bed_path = 'tests/fixtures/ccds-mini.bed'
    return _bed_path


@pytest.fixture
def sambamba_path():
    _sambamba_path = 'tests/fixtures/sambamba.depth.bed'
    return _sambamba_path



# -*- coding: utf-8 -*-
import pytest

from chanjo.sambamba import run_sambamba

THRESHOLDS = (10, 20)


def test_run_sambamba(tmpdir, bed_path, bam_path):
    out_path = tmpdir.join('ccds.coverage.bed')
    run_sambamba(bam_path, bed_path, outfile=str(out_path),
                 cov_treshold=THRESHOLDS)
    assert out_path.exists()


def test_run_sambamba_missing(tmpdir, reset_path, bed_path, bam_path):
    out_path = tmpdir.join('ccds.coverage.bed')
    with pytest.raises(OSError):
        run_sambamba(bam_path, bed_path, outfile=str(out_path),
                     cov_treshold=THRESHOLDS)



# -*- coding: utf-8 -*-
import codecs
import logging

import click
from path import path
import yaml
from distutils.spawn import find_executable

from chanjo.store.api import ChanjoDB
from .bootstrap import pull, BED_NAME, DB_NAME
from .demo import setup_demo

log = logging.getLogger(__name__)


@click.command()
@click.option('-f', '--force', is_flag=True, help='overwrite existing files')
@click.option('-d', '--demo', is_flag=True, help='copy demo files')
@click.option('-a', '--auto', is_flag=True)
@click.argument('root_dir', default='.', required=False)
@click.pass_context
def init(context, force, demo, auto, root_dir):
    """Bootstrap a new chanjo setup."""
    root_path = path(root_dir).abspath()
    log.info("setting up chanjo under: %s", root_path)
    db_uri = context.obj.get('database')
    abs_db_path = root_path.joinpath(DB_NAME)
    db_uri = db_uri or "sqlite:///{}".format(abs_db_path)
    # test setup of sambamba
    sambamba_bin = find_executable('sambamba')
    if sambamba_bin is None:  # pragma: no cover
        log.warn("'sambamba' command not found")
    else:
        log.debug("'sambamba' found: %s", sambamba_bin)

    if demo:
        log.info("copying demo files: %s", root_dir)
        setup_demo(root_dir, force=force)
    elif auto or click.confirm('Bootstrap CCDS transcript BED?'):
        # ensure root dir exists
        root_path.makedirs_p()
        pull(root_dir, force=force)

        log.info("configure new chanjo database: %s", db_uri)
        chanjo_db = ChanjoDB(db_uri)
        chanjo_db.set_up()

    # setup config file
    conf_path = root_path.joinpath('chanjo.yaml')
    with codecs.open(conf_path, 'w', encoding='utf-8') as conf_handle:
        data = {'database': db_uri}
        data_str = yaml.dump(data, default_flow_style=False)
        log.info("writing config file: %s", conf_path)
        conf_handle.write(data_str)

    click.echo('Chanjo bootstrap successful! Now run: ')
    bed_path = root_path.joinpath(BED_NAME)
    click.echo("chanjo --config {} link {}".format(conf_path, bed_path))



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import os
from mock import patch
from chanjo.init import bootstrap

from test_utils import FakeZipFile, fake_urlretrieve


@patch('chanjo.compat.urlretrieve', fake_urlretrieve)
@patch('zipfile.ZipFile', FakeZipFile)
def test_pull(tmpdir):
    # GIVEN a target directory
    target_dir = str(tmpdir)
    # WHEN downloading resources
    bootstrap.pull(target_dir)
    # THEN BED resource should be in place
    out_bed = str(tmpdir.join(bootstrap.BED_NAME))
    assert os.path.exists(out_bed)
    assert len(tmpdir.listdir()) == 1

    # GIVEN the resources already exists
    bootstrap.pull(target_dir)
    # THEN nothing happens :)



# -*- coding: utf-8 -*-
from chanjo.calculate import utils


def test_dump_json():
    # GIVEN some dict
    data = {'name': 'PT Anderson', 'age': 45}
    # WHEN dumping to JSON with pretty-option enabled
    json = utils.dump_json(data, pretty=True)
    # THEN the output is formatted over multiple lines
    assert isinstance(json, str)
    assert len(json.split('\n')) == 4



# -*- coding: utf-8 -*-
from __future__ import division
from collections import namedtuple
import logging
import subprocess

logger = logging.getLogger(__name__)

SexGuess = namedtuple('SexGuess', ['x_coverage', 'y_coverage', 'sex'])


def predict_sex(x_coverage, y_coverage):
    """Make a simple yet accurate prediction of a samples sex.

    The calculation is based on the average coverage across the X and Y
    chromosomes. Note that an extrapolation from a subsection of bases is
    usually quite sufficient.

    Args:
        x_coverage (float): estimated average coverage on X chromosome
        y_coverage (float): estimated average coverage on Y chromosome

    Returns:
        str: prediction, ['male', 'female', 'unknown']
    """
    # algoritm doesn't work if coverage is missing for X chromosome
    if x_coverage == 0:
        return 'unknown'
    elif (y_coverage > 0) and (x_coverage / y_coverage < 10):
        # this is the entire prediction, it's usually very obvious
        return 'male'
    else:
        # the few reads mapping to the Y chromosomes are artifacts
        return 'female'


def sex_from_bam(bam_path, prefix=''):
    """Predict the sex from a BAM alignment file.

    Args:
        bam_path (path): path to a BAM alignment file
        prefix (str, optional): string to prefix to 'X', 'Y'

    Returns:
        SexGuess: tuple of X coverage, Y coverage, and sex prediction

    Examples:
        >>> sex_from_bam('alignment.bam', prefix='chr')
        SexGuess(x_coverage=123.31, y_coverage=0.13, sex='female')
    """
    # make up some sex chromosome regions
    regions = ["{}X:1-59373566".format(prefix),
               "{}Y:69362-11375310".format(prefix)]
    averages = []
    for region in regions:
        command = ["sambamba depth region -L {} {}".format(region, bam_path)]
        logger.debug("calling: %s", command)
        bed_out = subprocess.check_output(command, shell=True).decode('utf-8')
        bed_rows = [line.split() for line in bed_out.splitlines()
                    if not line.startswith('#')]
        if len(bed_rows) == 1:
            averages.append(float(bed_rows[0][4]))
        else:
            chromosome = region.split(':')[0]
            logger.warn("couldn't find any reads on %s-chromosome", chromosome)
            averages.append(0.)

    # make the guess
    x_coverage, y_coverage = list(averages)
    sex = predict_sex(x_coverage, y_coverage)
    return SexGuess(x_coverage, y_coverage, sex)



# -*- coding: utf-8 -*-
from chanjo.store.models import TranscriptStat, Exon


def test_TranscriptStat():
    # GIVEN a transcript stat with two incomplete exons
    exons = [Exon('1', 10, 100, 99.1), Exon('1', 200, 300, 80.5)]
    stat = TranscriptStat(mean_coverage=10.3, incomplete_exons=exons)
    # WHEN accessing them
    parsed_exons = list(stat.incomplete_exons)
    # THEN should be the same
    assert parsed_exons == exons



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from tempfile import gettempdir

from mock import patch
import yaml

from test_utils import FakeZipFile, fake_urlretrieve


def test_init_demo(tmpdir, invoke_cli):
    # GIVEN empty directory
    assert tmpdir.listdir() == []
    # WHEN setting up demo on CLI
    target_dir = tmpdir.join('chanjo-demo')
    target_path = str(target_dir)
    result = invoke_cli(['init', '--demo', target_path])
    # THEN is should work and place 6 + 1 (config) in the folder
    assert result.exit_code == 0
    assert len(target_dir.listdir()) == 7


@patch('chanjo.compat.urlretrieve', fake_urlretrieve)
@patch('zipfile.ZipFile', FakeZipFile)
@patch('click.confirm', lambda param: True)
def test_init_bootstrap(tmpdir, invoke_cli):
    # GIVEN empty dir
    assert tmpdir.listdir() == []
    # WHEN boostrapping chanjo
    result = invoke_cli(['init', str(tmpdir)])
    # THEN it should place 2 + 1 (BED, DB, config) files in the target dir
    assert result.exit_code == 0
    assert len(tmpdir.listdir()) == 3


@patch('click.confirm', lambda param: False)
@patch('click.prompt', lambda param: "{}/cov.sqlite3".format(gettempdir()))
def test_init_no_bootstrap(tmpdir, invoke_cli):
    # GIVEN ...
    # WHEN opting out of bootstrap
    result = invoke_cli(['init', str(tmpdir)])
    # THEN it should work with custom database URI
    assert result.exit_code == 0
    conf_path = tmpdir.join('chanjo.yaml')
    with open(str(conf_path), 'r') as handle:
        data = yaml.load(handle)
    assert 'coverage.sqlite3' in data['database']



# -*- coding: utf-8 -*-
"""
chanjo
~~~~~~~
Coverage analysis for clinical sequencing.

:copyright: (c) 2014 by Robin Andeer
:licence: MIT, see LICENCE for more details
"""
import logging
from pkg_resources import get_distribution

__banner__ = r"""
       ______               ________
 _________  /_______ _____________(_)_____
 _  ___/_  __ \  __ `/_  __ \____  /_  __ \
 / /__ _  / / / /_/ /_  / / /___  / / /_/ /   by Robin Andeer
 \___/ /_/ /_/\__,_/ /_/ /_/___  /  \____/
                            /___/
"""

__title__ = 'chanjo'
__summary__ = 'coverage analysis tool for clinical sequencing'
__uri__ = 'http://www.chanjo.co/'

__version__ = get_distribution(__title__).version
__codename__ = 'Optimistic Otter'

__author__ = 'Robin Andeer'
__email__ = 'robin.andeer@gmail.com'

__license__ = 'MIT'
__copyright__ = 'Copyright 2016 Robin Andeer'

# the user should dictate what happens when a logging event occurs
logging.getLogger(__name__).addHandler(logging.NullHandler())



# -*- coding: utf-8 -*-
from chanjo.load import link


def test_process(bed_lines):
    result = link.link_elements(bed_lines)
    models = list(result.models)
    assert result.count == 5  # 5 transcripts



# -*- coding: utf-8 -*-
from chanjo.store.models import Sample


def test_mean(populated_db):
    # GIVEN a database loaded with one sample
    assert Sample.query.count() == 1
    # WHEN calculating mean values across metrics
    query = populated_db.mean()
    # THEN the results should group over one "row"
    results = query.all()
    assert len(results) == 1
    sample = results[0]
    assert sample[0] == 'sample'  # sample id
    for metric in filter(None, sample[1:]):
        assert isinstance(metric, float)


def test_gene(populated_db):
    # GIVEN a database populated with a single sample
    assert Sample.query.count() == 1
    # WHEN calculating average metrics for a gene
    query = populated_db.gene_metrics('SAMD11')
    # THEN the results should add up to a single row
    results = query.all()
    assert len(results) == 1
    sample = results[0]
    assert sample[0] == 'sample'
    assert sample[-1] == 'SAMD11'



# -*- coding: utf-8 -*-
import click
import logging

from chanjo.compat import text_type
from .core import sex_from_bam

log = logging.getLogger(__name__)


@click.command()
@click.option('-p', '--prefix', default='', help='chromosome prefix')
@click.argument('bam_path', type=click.Path(exists=True))
@click.pass_context
def sex(context, prefix, bam_path):
    """Guess the sex of a BAM alignment."""
    try:
        result = sex_from_bam(bam_path, prefix=prefix)
    except Exception:
        log.exception('Something went really wrong :(')
        context.abort()

    # print the results to the console for pipeability (csv)
    click.echo("#{prefix}X_coverage\t{prefix}Y_coverage\tsex"
               .format(prefix=prefix))
    click.echo('\t'.join(map(text_type, result)))



# -*- coding: utf-8 -*-
"""
chanjo.store.core
~~~~~~~~~~~~~~~~~~
"""
from __future__ import division
import logging
import os
import types

from alchy import Manager

from chanjo.calculate.mixins import CalculateMixin
from .models import BASE, TranscriptStat

log = logging.getLogger(__name__)


class ChanjoDB(Manager, CalculateMixin):

    """SQLAlchemy-based database object.

    Bundles functionality required to setup and interact with various
    related genomic interval elements.

    .. versionchanged:: 2.1.0
        Lazy-loadable, all "init" arguments optional.

    Examples:
        >>> chanjo_db = Store('data/elements.sqlite3')
        >>> chanjo_db.set_up()

    .. note::

        For testing pourposes use ``:memory:`` as the ``path`` argument to
        set up in-memory (temporary) database.

    Args:
        uri (Optional[str]): path/URI to the database to connect to
        debug (Optional[bool]): whether to output logging information
        base (Optional[sqlalchemy.ext.declarative.api.Base]): schema definition

    Attributes:
        uri (str): path/URI to the database to connect to
        base (sqlalchemy.ext.declarative.api.Base): shcema definition
        engine (class): SQLAlchemy engine, defines what database to use
        session (class): SQLAlchemy ORM session, manages persistance
        query (method): SQLAlchemy ORM query builder method
        classes (dict): bound ORM classes
    """

    def __init__(self, uri=None, debug=False, base=BASE):
        self.Model = base
        self.uri = uri
        if uri:
            self.connect(uri, debug=debug)

    def connect(self, db_uri, debug=False):
        """Configure connection to a SQL database.

        .. versionadded:: 2.1.0

        Args:
            db_uri (str): path/URI to the database to connect to
            debug (Optional[bool]): whether to output logging information
        """
        config = {'SQLALCHEMY_ECHO': debug}
        if 'mysql' in db_uri:  # pragma: no cover
            config['SQLALCHEMY_POOL_RECYCLE'] = 3600
        elif '://' not in db_uri:
            # expect only a path to a sqlite database
            db_path = os.path.abspath(os.path.expanduser(db_uri))
            db_uri = "sqlite:///{}".format(db_path)
            self.uri = db_uri

        config['SQLALCHEMY_DATABASE_URI'] = db_uri

        # connect to the SQL database
        super(ChanjoDB, self).__init__(config=config, Model=self.Model)

        # shortcut to query method
        self.query = self.session.query
        return self

    @property
    def dialect(self):
        """Return database dialect name used for the current connection.

        Dynamic attribute.

        Returns:
            str: name of dialect used for database connection
        """
        return self.engine.dialect.name

    def set_up(self):
        """Initialize a new database with the default tables and columns.

        Returns:
            Store: self
        """
        # create the tables
        self.create_all()
        tables = self.Model.metadata.tables.keys()
        log.info("created tables: %s", ', '.join(tables))
        return self

    def tear_down(self):
        """Tear down a database (tables and columns).

        Returns:
            Store: self
        """
        # drop/delete the tables
        self.drop_all()
        return self

    def save(self):
        """Manually persist changes made to various elements. Chainable.

        .. versionchanged:: 2.1.2
            Flush session before commit.

        Returns:
            Store: ``self`` for chainability
        """
        try:
            # commit/persist dirty changes to the database
            self.session.flush()
            self.session.commit()
        except Exception as error:
            log.debug('rolling back failed transaction')
            self.session.rollback()
            raise error
        return self

    def add(self, items):
        """Add one or more new items and commit the changes. Chainable.

        Args:
            items (orm/list): new ORM object instance or list of such

        Returns:
            Store: ``self`` for chainability
        """
        if isinstance(items, self.Model):
            # Add the record to the session object
            self.session.add(items)
        elif isinstance(items, list):
            # Add all records to the session object
            self.session.add_all(items)
        elif isinstance(items, types.GeneratorType):
            # Iterate over all items
            for element in items:
                self.session.add(element)
        else:
            raise ValueError('unknown object type')
        return self



# -*- coding: utf-8 -*-
import logging

LEVELS = {
    0: 'WARNING',
    1: 'INFO',
    2: 'DEBUG',
}


def init_log(logger, filename=None, loglevel=None):
    """Initializes the log file in the proper format.

    Arguments:
        filename (str): Path to a file. Or None if logging is to
                         be disabled.
        loglevel (str): Determines the level of the log output.
    """
    template = "[%(asctime)s] %(levelname)-8s: %(name)-25s: %(message)s"
    formatter = logging.Formatter(template)

    if loglevel:
        logger.setLevel(getattr(logging, loglevel))

    # We will always print warnings and higher to stderr
    console = logging.StreamHandler()
    console.setLevel('INFO')
    console.setFormatter(formatter)

    if filename:
        file_handler = logging.FileHandler(filename, encoding='utf-8')
        if loglevel:
            file_handler.setLevel(getattr(logging, loglevel))
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
    # If no logfile is provided we print all log messages that the user has
    # defined to stderr
    else:
        if loglevel:
            console.setLevel(getattr(logging, loglevel))

    logger.addHandler(console)



# -*- coding: utf-8 -*-
"""Component for guessing the sex from a BAM alignment.

The component reads coverage for subsections of each sex chromosome.
Based on the ratio between the average coverage across chromosomes it
makes a simple sex prediction.
"""
from .core import predict_sex, sex_from_bam, SexGuess
from .cli import sex as sex_cmd



# -*- coding: utf-8 -*-
from sqlalchemy.sql import func

from chanjo.store.models import Transcript, TranscriptStat


class CalculateMixin:

    """Methods for calculating various metrics."""

    def mean(self, sample_ids=None):
        """Calculate the mean values of all metrics per sample."""
        sql_query = (self.query(TranscriptStat.sample_id,
                                func.avg(TranscriptStat.mean_coverage),
                                func.avg(TranscriptStat.completeness_10),
                                func.avg(TranscriptStat.completeness_15),
                                func.avg(TranscriptStat.completeness_20),
                                func.avg(TranscriptStat.completeness_50),
                                func.avg(TranscriptStat.completeness_100))
                         .group_by(TranscriptStat.sample_id))
        return sql_query

    def gene_metrics(self, gene):
        """Calculate gene statistics."""
        query = (self.mean()
                     .add_column(Transcript.gene_id)
                     .join(TranscriptStat.transcript)
                     .filter(Transcript.gene_id == gene)
                     .group_by(Transcript.gene_id))
        return query



# -*- coding: utf-8 -*-
"""
chanjo.compat
~~~~~~~~~~~~~~~

Python 2.7.x, 3.2+ compatability module.
"""
import operator
import sys

is_py2 = sys.version_info[0] == 2


if not is_py2:  # pragma: no cover
    # Python 3
    from urllib.request import urlretrieve

    # strings and ints
    text_type = str
    string_types = (str,)
    integer_types = (int,)

    # lazy iterators
    zip = zip
    range = range
    iteritems = operator.methodcaller('items')
    iterkeys = operator.methodcaller('keys')
    itervalues = operator.methodcaller('values')

else:
    # Python 2
    from urllib import urlretrieve

    # strings and ints
    text_type = unicode
    string_types = (str, unicode)
    integer_types = (int, long)

    # lazy iterators
    range = xrange
    from itertools import izip as zip
    iteritems = operator.methodcaller('iteritems')
    iterkeys = operator.methodcaller('iterkeys')
    itervalues = operator.methodcaller('itervalues')



# -*- coding: utf-8 -*-
from chanjo.load import utils


def test_groupby_tx(bed_exons, sambamba_exons):
    transcripts = list(utils.groupby_tx(bed_exons))
    assert len(transcripts) == 5

    # GIVEN sambamba lines
    transcripts = list(utils.groupby_tx(sambamba_exons, sambamba=True))
    assert len(transcripts) == 10



# -*- coding: utf-8 -*-
from __future__ import division
from collections import namedtuple

from chanjo.compat import iteritems, itervalues
from chanjo.store.models import TranscriptStat, Sample, Exon
from .parse import sambamba
from .utils import groupby_tx

Result = namedtuple('Result', ['models', 'count', 'sample'])


def load_transcripts(sequence, sample_id=None, group_id=None, source=None,
                     threshold=None):
    """Process a sequence of exon lines.

    Args:
        sequence (sequence): list of chanjo bed lines
        sample_id (Optional[str]): unique sample id, else auto-guessed
        grouip_id (Optional[str]): id to group samples
        source (Optional[str]): path to coverage source (BAM/Sambamba)
        threshold (Optional[int]): completeness level to disqualify exons

    Returns:
        Result: iterators of `Transcript`, transcripts processed, sample model
    """
    exons = sambamba.depth_output(sequence)
    transcripts = groupby_tx(exons, sambamba=True)
    raw_stats = ((tx_id, tx_stat(tx_id, exons, threshold=threshold))
                 for tx_id, exons in iteritems(transcripts))

    if sample_id is None:
        sample_id = next(iter(itervalues(transcripts)))[0]['sampleName']
    sample_obj = Sample(id=sample_id, group_id=group_id, source=source)

    models = (make_model(sample_obj, tx_id, raw_stat) for tx_id, raw_stat
              in raw_stats)
    return Result(models=models, count=len(transcripts), sample=sample_obj)


def tx_stat(transcript_id, exons, threshold=None):
    """Calculate metrics for transcript stats model.

    Args:
        transcript_id (str): unqiue transcript id
        exons (List[dict]): list of exon transcripts
        threshold (Optional[int]): completeness level to disqualify exons

    Returns:
        dict: aggregated stats over all exons
    """
    sums = {'bases': 0, 'mean_coverage': 0}
    incomplete_exons = []

    # for each of the exons (linked to one transcript)
    for exon in exons:
        # go over each of the fields to sum up
        exon_length = (exon['chromEnd'] - exon['chromStart'])
        sums['bases'] += exon_length
        sums['mean_coverage'] += (exon['meanCoverage'] * exon_length)

        # add to the total sum for completeness levels
        for comp_key in [10, 15, 20, 50, 100]:
            if comp_key in exon['thresholds']:
                sums_key = "completeness_{}".format(comp_key)
                if sums_key not in sums:
                    sums[sums_key] = 0
                completeness = exon['thresholds'][comp_key]
                sums[sums_key] += (completeness * exon_length)

                if threshold == comp_key and completeness < 100:
                    exon_obj = Exon(exon['chrom'], exon['chromStart'],
                                    exon['chromEnd'], completeness)
                    incomplete_exons.append(exon_obj)

    fields = {key: (value / sums['bases']) for key, value in iteritems(sums)
              if key != 'bases'}
    fields['incomplete_exons'] = incomplete_exons
    fields['threshold'] = threshold
    return fields


def make_model(sample_obj, transcript_id, fields):
    """Compose a transcript stat model from fields.

    Args:
        sample_obj (Sample): Sample database model
        transcript_id (str): unique transcript id
        fields (dict): key/values of metrics

    Returns:
        Transcript: composed transcript model
    """
    tx_model = TranscriptStat(sample_id=sample_obj.id,
                              transcript_id=transcript_id, **fields)
    return tx_model



# -*- coding: utf-8 -*-
from errno import EEXIST
import logging
from pkg_resources import resource_filename, resource_listdir

from path import path

log = logging.getLogger(__name__)


def setup_demo(location, force=False):
    """Copy demo files to a directory.

    \b
    LOCATION: directory to add demofiles to (default: ./chanjo-demo)
    """
    target_dir = path(location)
    pkg_dir = __name__.rpartition('.')[0]
    demo_dir = path(resource_filename(pkg_dir, 'demo-files'))

    # make sure we don't overwrite exiting files
    for demo_file in resource_listdir(pkg_dir, 'demo-files'):
        target_file_path = target_dir.joinpath(demo_file)
        if not force and target_file_path.exists():
            log.error("%s exists, pick a different location", target_file_path)
            raise OSError(EEXIST, 'file already exists', target_file_path)

    try:
        # we can copy the directory(tree)
        demo_dir.copytree(target_dir)
    except OSError as error:
        log.warn('location must be a non-existing directory')
        raise error

    # inform the user
    log.info("successfully copied demo files to %s", target_dir)



# -*- coding: utf-8 -*-
import logging

import click

from chanjo.compat import zip
from chanjo.store.api import ChanjoDB
from chanjo.store.constants import STAT_COLUMNS
from .utils import dump_json

log = logging.getLogger(__name__)


@click.group()
@click.pass_context
def calculate(context):
    """Calculate statistics across samples."""
    context.obj['db'] = ChanjoDB(uri=context.obj['database'])


@calculate.command()
@click.option('-p', '--pretty', is_flag=True)
@click.pass_context
def mean(context, pretty):
    """Calculate mean statistics."""
    query = context.obj['db'].mean()
    columns = ['sample_id'] + STAT_COLUMNS
    for sample in query:
        result = {column: value for column, value in zip(columns, sample)}
        click.echo(dump_json(result, pretty=pretty))


@calculate.command()
@click.option('-p', '--pretty', is_flag=True)
@click.argument('gene')
@click.pass_context
def gene(context, pretty, gene):
    """Calculate stats for a given gene."""
    query = context.obj['db'].gene_metrics(gene)
    columns = ['sample_id'] + STAT_COLUMNS + ['gene_id']
    for sample in query:
        result = {column: value for column, value in zip(columns, sample)}
        click.echo(dump_json(result, pretty=pretty))



# -*- coding: utf-8 -*-
from collections import namedtuple
from datetime import datetime

from alchy import ModelBase, make_declarative_base
from sqlalchemy import Column, types, ForeignKey, UniqueConstraint, orm

Exon = namedtuple('Exon', ['chrom', 'start', 'end', 'completeness'])

# base for declaring a mapping
BASE = make_declarative_base(Base=ModelBase)


class Transcript(BASE):

    """Set of non-overlapping exons.

    A :class:`Transcript` can *only* be related to a single gene.

    Args:
        id (str): unique transcript id (e.g. CCDS)
        gene_id (str): related gene
        chromosome (str): related contig id
        lenght (int): number of exon bases in transcript
    """

    __tablename__ = 'transcript'

    id = Column(types.String(32), primary_key=True)
    gene_id = Column(types.String(32), index=True, nullable=False)
    chromosome = Column(types.String(10))
    length = Column(types.Integer)

    stats = orm.relationship('TranscriptStat', backref='transcript')


class Sample(BASE):

    """Metadata for a single sample.

    Args:
        id (str): unique sample id
        group_id (str): unique group id
        source (str): path to coverage source Sambamba output/BAM file
        created_at (DateTime): date of addition to database
    """

    __tablename__ = 'sample'

    id = Column(types.String(32), primary_key=True)
    group_id = Column(types.String(32), index=True)
    source = Column(types.String(256))
    created_at = Column(types.DateTime, default=datetime.now)

    sample = orm.relationship('TranscriptStat', cascade='all,delete',
                              backref='sample')


class TranscriptStat(BASE):

    """Statistics on transcript level, related to sample and transcript.

    Args:
        sample_id (str): link to sample record
        sample (Sample): parent Sample record
        transcript_id (str): link to transcript record
        transcript (Transcript): parent transcript record
        mean_coverage (Float): mean coverage across all exons
        completeness_XX (Float): percentage of exon bases coverage at XX
        _incomplete_exons (str): comma separated list of exon ids
    """

    __tablename__ = 'transcript_stat'
    __table_args__ = (UniqueConstraint('sample_id', 'transcript_id',
                                       name='_sample_transcript_uc'),)

    id = Column(types.Integer, primary_key=True)
    mean_coverage = Column(types.Float, nullable=False)
    completeness_10 = Column(types.Float)
    completeness_15 = Column(types.Float)
    completeness_20 = Column(types.Float)
    completeness_50 = Column(types.Float)
    completeness_100 = Column(types.Float)

    threshold = Column(types.Integer)
    _incomplete_exons = Column(types.Text)

    sample_id = Column(types.String(32), ForeignKey('sample.id'),
                       nullable=False)
    transcript_id = Column(types.String(32), ForeignKey('transcript.id'),
                           nullable=False)

    @property
    def incomplete_exons(self):
        """Return a list of exons ids."""
        raw_exons = (self._incomplete_exons.split(',') if
                     self._incomplete_exons else [])
        for raw_exon in raw_exons:
            data = raw_exon.split('|')
            yield Exon(chrom=data[0], start=int(data[1]), end=int(data[2]),
                       completeness=float(data[3]))

    @incomplete_exons.setter
    def incomplete_exons(self, exon_list):
        raw_exons = ['|'.join(map(str, exon)) for exon in exon_list]
        self._incomplete_exons = ','.join(raw_exons) if raw_exons else None



# -*- coding: utf-8 -*-
import logging

import click

from .api import ChanjoDB
from .models import Sample, TranscriptStat

log = logging.getLogger(__name__)


@click.group()
@click.pass_context
def db(context):
    """Interact with the database for maintainance tasks."""
    context.obj['db'] = ChanjoDB(uri=context.obj['database'])


@db.command()
@click.option('--reset', is_flag=True, help='tear down existing db')
@click.pass_context
def setup(context, reset):
    """Initialize a new datbase from scratch."""
    if reset:
        log.info('tearing down existing database')
        context.obj['db'].tear_down()
    log.info('setting up new database')
    context.obj['db'].set_up()


@db.command()
@click.argument('sample_id', type=str)
@click.pass_context
def remove(context, sample_id):
    """Remove all traces of a sample from the database."""
    db = context.obj['db']
    log.debug('find sample in database with id: %s', sample_id)
    sample_obj = Sample.query.get(sample_id)
    if sample_obj is None:
        log.warn('sample (%s) not found in database', sample_id)
        context.abort()
    log.info('delete sample (%s) from database', sample_id)
    db.session.delete(sample_obj)
    db.save()



# -*- coding: utf-8 -*-
import json

from chanjo.cli import root
from chanjo.store.models import Sample


def test_mean(popexist_db, cli_runner):
    # GIVEN an existing databse with one sample
    assert Sample.query.count() == 1
    # WHEN assessing the mean values metrics per sample
    res = cli_runner.invoke(root, ['-d', popexist_db.uri, 'calculate', 'mean'])
    # THEN the command should return JSON results
    assert res.exit_code == 0
    # ... returns some debug info to STDERR that we strip away
    lines = res.output.strip().split('\n')
    assert len(lines) == 2
    # ... the last row (no incl. empty line) is JSON formatted
    data = json.loads(lines[1].strip())
    assert data['sample_id'] == 'sample'
    assert isinstance(data['mean_coverage'], float)


def test_gene(popexist_db, cli_runner):
    # GIVEN an existing databse with one sample
    sample_id = 'sample'
    gene_id = 'WDR85'
    assert Sample.query.first().id == sample_id
    # WHEN assessing the metrics for a gene (and sample)
    res = cli_runner.invoke(root, ['-d', popexist_db.uri, 'calculate', 'gene',
                            gene_id])
    # THEN the command should return JSON results
    assert res.exit_code == 0
    # ... returns some debug info to STDERR that we strip away
    lines = res.output.strip().split('\n')
    assert len(lines) == 2
    # ... the last row (no incl. empty line) is JSON formatted
    data = json.loads(lines[1].strip())
    assert data['sample_id'] == sample_id
    assert data['gene_id'] == gene_id
    assert isinstance(data['mean_coverage'], float)



# -*- coding: utf-8 -*-
import os

import pytest

from chanjo.init import demo


def test_setup_demo(tmpdir):
    # GIVEN a non-existing target directory
    target_dir = tmpdir.join('chanjo-demo')
    target_path = str(target_dir)
    demo.setup_demo(target_path)
    files = os.listdir(target_path)
    expected_files = os.listdir('chanjo/init/demo-files')
    assert files == expected_files

    # GIVEN the files are already in place
    with pytest.raises(OSError):
        demo.setup_demo(target_path)


def test_setup_demo_existing_dir(tmpdir):
    with pytest.raises(OSError):
        demo.setup_demo(str(tmpdir))



# -*- coding: utf-8 -*-
"""
chanjo.cli
~~~~~~~~~~~

Command line interface (console entry points). Based on Click_.

.. _Click: http://click.pocoo.org/
"""
from datetime import date
import logging
import os
import pkg_resources

import click
import yaml

from chanjo import __version__, __title__
from chanjo.compat import text_type
from chanjo.log import init_log

logger = logging.getLogger(__name__)
TODAY = str(date.today())


class EntryPointsCLI(click.MultiCommand):

    """Add subcommands dynamically to a CLI via entry points."""

    def _iter_commands(self):
        """Iterate over all subcommands as defined by the entry point."""
        return {entry_point.name: entry_point for entry_point in
                pkg_resources.iter_entry_points('chanjo.subcommands.4')}

    def list_commands(self, ctx):
        """List the available commands."""
        commands = self._iter_commands()
        return commands.keys()

    def get_command(self, ctx, name):
        """Load one of the available commands."""
        commands = self._iter_commands()
        if name not in commands:
            click.echo("no such command: {}".format(name))
            ctx.abort()
        return commands[name].load()


@click.group(cls=EntryPointsCLI)
@click.option('-c', '--config', default='~/.chanjo.yaml',
              type=click.Path(), help='path to config file')
@click.option('-d', '--database', type=text_type,
              help='path/URI of the SQL database')
@click.option('-l', '--log-level', default='INFO')
@click.option('--log-file', type=click.Path())
@click.version_option(__version__, prog_name=__title__)
@click.pass_context
def root(context, config, database, log_level, log_file):
    """Clinical sequencing coverage analysis tool."""
    init_log(logging.getLogger(), loglevel=log_level, filename=log_file)
    logger.info("version {0}".format(__version__))

    # avoid setting global defaults in Click options, do it below when
    if os.path.exists(config):
        with open(config) as conf_handle:
            context.obj = yaml.load(conf_handle)
    else:
        context.obj = {}
    context.obj['database'] = (database or context.obj.get('database'))

    # update the context with new defaults from the config file
    context.default_map = context.obj



# -*- coding: utf-8 -*-
import os

import yaml


def test_logging_to_file(tmpdir, invoke_cli):
    # GIVEN an empty directory
    assert tmpdir.listdir() == []
    # WHEN running the CLI to display some help for a subcommand
    log_path = tmpdir.join('stderr.log')
    result = invoke_cli(['--log-file', str(log_path), 'db'])
    assert result.exit_code == 0
    assert tmpdir.listdir() == [log_path]


def test_list_commands(invoke_cli):
    # WHEN using simplest command 'chanjo'
    result = invoke_cli()
    # THEN it should just work ;-)
    assert result.exit_code == 0


def test_missing_command(invoke_cli):
    # WHEN invoking a missing command
    result = invoke_cli(['idontexist'])
    # THEN context should abort
    assert result.exit_code != 0


def test_with_config(tmpdir, invoke_cli):
    # GIVEN a simple config file
    conf_path = str(tmpdir.join('chanjo.yaml'))
    db_path = str(tmpdir.join('coverage.sqlite3'))
    data = {'database': db_path}
    with open(conf_path, 'w') as handle:
        handle.write(yaml.dump(data))
    # WHEN launching CLI with config
    result = invoke_cli(['-c', conf_path, 'db', 'setup'])
    # THEN config values should be picked up
    assert result.exit_code == 0
    assert os.path.exists(db_path)



# -*- coding: utf-8 -*-
COMPLETENESS_LEVELS = [10, 15, 20, 50, 100]
COMPLETENESS_COLUMNS = ["completeness_{}".format(level) for level
                        in COMPLETENESS_LEVELS]
STAT_COLUMNS = ['mean_coverage'] + COMPLETENESS_COLUMNS



# -*- coding: utf-8 -*-
import logging
from collections import namedtuple

from chanjo.compat import iteritems
from chanjo.store.models import Transcript
from .parse import bed as parse_bed
from .utils import groupby_tx

Result = namedtuple('Result', ['models', 'count'])
log = logging.getLogger(__name__)


def link_elements(sequence):
    """Process a sequence of exon lines.

    Args:
        sequence (sequence): list of chanjo bed lines

    Returns:
        Result: iterators of transcript models, number of transcripts processed
    """
    exons = parse_bed.chanjo(sequence)
    transcripts = groupby_tx(exons)
    models = (make_model(tx_id, exons) for tx_id, exons in
              iteritems(transcripts))
    return Result(models=models, count=len(transcripts))


def make_model(transcript_id, exons):
    """Generate transcript model from a list of exons.

    Args:
        transcript_id (str): unique transcript id
        exons (List[dict]): list of exon dictionaries

    Returns:
        Transcript: uncommitted transcript model
    """
    # assume the same chromosome and gene for all exons
    chromosome = exons[0]['chrom']
    gene_id = exons[0]['elements'][transcript_id]
    tot_length = sum((exon['chromEnd'] - exon['chromStart']) for exon in exons)
    tx_model = Transcript(id=transcript_id, chromosome=chromosome,
                          length=tot_length, gene_id=gene_id)
    return tx_model



# -*- coding: utf-8 -*-
import os.path
import logging

import click
from sqlalchemy.exc import IntegrityError

from chanjo.store.api import ChanjoDB
from .utils import validate_stdin
from .link import link_elements
from .sambamba import load_transcripts

log = logging.getLogger(__name__)


@click.command()
@click.option('-s', '--sample', help='override sample id from file')
@click.option('-g', '--group', help='id to group related samples')
@click.option('-r', '--threshold', type=int,
              help='completeness level to disqualify exons')
@click.argument('bed_stream', callback=validate_stdin,
                type=click.File(encoding='utf-8'), default='-', required=False)
@click.pass_context
def load(context, sample, group, threshold, bed_stream):
    """Load Sambamba output into the database for a sample."""
    chanjo_db = ChanjoDB(uri=context.obj['database'])
    source = os.path.abspath(bed_stream.name)

    result = load_transcripts(bed_stream, sample_id=sample, group_id=group,
                              source=source, threshold=threshold)
    try:
        chanjo_db.add(result.sample)
        with click.progressbar(result.models, length=result.count,
                               label='loading transcripts') as bar:
            for tx_model in bar:
                chanjo_db.session.add(tx_model)
        chanjo_db.save()
    except IntegrityError as error:
        log.error('sample already loaded, rolling back')
        log.debug(error.message)
        chanjo_db.session.rollback()
        context.abort()


@click.command()
@click.argument('bed_stream', callback=validate_stdin,
                type=click.File(encoding='utf-8'), default='-', required=False)
@click.pass_context
def link(context, bed_stream):
    """Link related genomic elements."""
    chanjo_db = ChanjoDB(uri=context.obj['database'])
    result = link_elements(bed_stream)
    with click.progressbar(result.models, length=result.count,
                           label='adding transcripts') as bar:
        for tx_model in bar:
            chanjo_db.add(tx_model)
    try:
        chanjo_db.save()
    except IntegrityError:
        click.echo('elements already linked, rolling back')
        chanjo_db.session.rollback()
        click.echo("use 'chanjo db setup --reset' to re-build")
        context.abort()



# -*- coding: utf-8 -*-
from chanjo.sex import SexGuess, sex_from_bam, predict_sex


def test_SexGuess():
    # create a sex guess tuple
    sex = SexGuess(10.2, 8.1, 'male')

    assert sex.x_coverage == 10.2
    assert sex.y_coverage == 8.1
    assert sex.sex == 'male'


def test_predict_sex():
    assert predict_sex(x_coverage=10.2, y_coverage=8.1) == 'male'
    assert predict_sex(x_coverage=20.9, y_coverage=.01) == 'female'
    # shouldn't raise ``ZeroDivisionError``
    assert predict_sex(x_coverage=5.12, y_coverage=0) == 'female'
    # what does this even mean?
    assert predict_sex(x_coverage=0, y_coverage=10) == 'unknown'


def test_sex_from_bam(bam_path):
    # use fixtures bam - doesn't have coverage on Y chromosome
    result = sex_from_bam(bam_path)
    assert result.x_coverage > result.y_coverage
    assert result.sex == 'female'



# -*- coding: utf-8 -*-
import json


def dump_json(data, pretty=False):
    """Print JSON to console."""
    if pretty:
        json_args = dict(indent=4, sort_keys=True)
    else:
        json_args = {}
    return json.dumps(data, **json_args)



# -*- coding: utf-8 -*-
import logging
import zipfile

from path import path

from chanjo.compat import urlretrieve

DB_NAME = 'chanjo.coverage.sqlite3'
BED_NAME = 'ccds.15.grch37p13.extended.bed'
BED_URL = ("https://s3.eu-central-1.amazonaws.com/"
           "chanjo/ccds15.grch37p13.extended.bed.zip")

logger = logging.getLogger(__name__)


def pull(target_dir, force=False):  # pragma: no cover
    """Download precompiled resources into a folder.

    Args:
        target_dir (path): relative path the folder to download to
        force (Optional[bool]): whether to overwrite existing files
    """
    logger.debug('ensure target directory exists')
    target_path = path(target_dir)
    target_path.makedirs_p()

    bed_zip_path = target_path.joinpath("{}.zip".format(BED_NAME))
    final_bed = target_path.joinpath(BED_NAME)

    if not final_bed.exists() or force:
        logger.info('downloading... [%s]', BED_URL)
        urlretrieve(BED_URL, bed_zip_path)

        logger.info('extracting BED file...')
        zip_ref = zipfile.ZipFile(bed_zip_path, 'r')
        zip_ref.extractall(target_dir)

        logger.info('removing BED archive...')
        bed_zip_path.remove_p()
    else:
        logger.warn('file already exists, skipping: %s', final_bed)



# -*- coding: utf-8 -*-
from chanjo.store.models import Sample, Transcript


def test_load(existing_db, invoke_cli, sambamba_path):
    # GIVEN processed sambamba depth output and empty database
    assert Sample.query.count() == 0
    db_uri = existing_db.uri
    result = invoke_cli(['--database', db_uri, 'load', sambamba_path])
    # WHEN loading into database
    assert result.exit_code == 0
    assert Sample.query.count() == 1


def test_load_conflict(popexist_db, invoke_cli, sambamba_path):
    # GIVEN an existing database with a sample
    db_uri = popexist_db.uri
    sample_id = 'sample'
    assert Sample.query.first().id == sample_id
    # WHEN you try to add the same sample id twice
    result = invoke_cli(['--database', db_uri, 'load', '--sample', sample_id,
                         sambamba_path])
    # THEN it should fail
    assert result.exit_code != 0
    assert Sample.query.count() == 1


def test_link(existing_db, invoke_cli, bed_path):
    # GIVEN chanjo bed file and an existing database
    db_uri = existing_db.uri
    # WHEN linking elements in the database
    result = invoke_cli(['--database', db_uri, 'link', bed_path])
    # THEN database should be populated with all transcripts
    assert result.exit_code == 0
    assert Transcript.query.count() == 5

    # WHEN loading again...
    result = invoke_cli(['--database', db_uri, 'link', bed_path])
    # THEN it should error and rollback the session
    assert result.exit_code != 0
    assert Transcript.query.count() == 5


# def test_load_with_stdin(invoke_cli):
#     # GIVEN the STDIN is empty
#     # WHEN loading data
#     result = invoke_cli(['--database', 'sqlite://', 'load'])
#     # THEN it should complain
#     assert result.exit_code != 0
#     assert result.exception == click.BadParameter



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Based on https://github.com/pypa/sampleproject/blob/master/setup.py."""
# To use a consistent encoding
import codecs
import io
import os
# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import sys

# Shortcut for building/publishing to Pypi
if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist bdist_wheel upload')
    sys.exit()


def parse_reqs(req_path='./requirements.txt'):
    """Recursively parse requirements from nested pip files."""
    install_requires = []
    with codecs.open(req_path, 'r') as handle:
        # remove comments and empty lines
        lines = (line.strip() for line in handle
                 if line.strip() and not line.startswith('#'))
        for line in lines:
            # check for nested requirements files
            if line.startswith('-r'):
                # recursively call this function
                install_requires += parse_reqs(req_path=line[3:])
            else:
                # add the line as a new requirement
                install_requires.append(line)
    return install_requires


# This is a plug-in for setuptools that will invoke py.test
# when you run python setup.py test
class PyTest(TestCommand):

    """Set up the py.test test runner."""

    def finalize_options(self):
        """Set options for the command line."""
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        """Execute the test runner command."""
        # Import here, because outside the required eggs aren't loaded yet
        import pytest
        sys.exit(pytest.main(self.test_args))

# Get the long description from the relevant file
here = os.path.abspath(os.path.dirname(__file__))
with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='chanjo',

    # Versions should comply with PEP440. For a discussion on
    # single-sourcing the version across setup.py and the project code,
    # see http://packaging.python.org/en/latest/tutorial.html#version
    version='4.0.0-beta',

    description='Coverage analysis tool for clinical sequencing',
    long_description=long_description,
    # What does your project relate to? Separate with spaces.
    keywords='coverage sequencing clinical exome completeness diagnostics',
    author='Robin Andeer',
    author_email='robin.andeer@gmail.com',
    license='MIT',

    # The project's main homepage
    url='http://www.chanjo.co/',

    packages=find_packages(exclude=('tests*', 'docs', 'examples')),

    # If there are data files included in your packages that need to be
    # installed, specify them here.
    include_package_data=True,
    package_data=dict(chanjo=['demo/files/*']),
    zip_safe=False,

    # Install requirements loaded from ``requirements.txt``
    install_requires=parse_reqs(),
    tests_require=[
        'pytest',
    ],
    cmdclass=dict(test=PyTest),

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and
    # allow pip to create the appropriate form of executable for the
    # target platform.
    entry_points={
        'console_scripts': [
            'chanjo = chanjo.cli:root',
        ],
        'chanjo.subcommands.4': [
            'init = chanjo.init.cli:init',
            'sex = chanjo.sex.cli:sex',
            'sambamba = chanjo.sambamba.cli:sambamba',
            'db = chanjo.store.cli:db',
            'load = chanjo.load.cli:load',
            'link = chanjo.load.cli:link',
            'calculate = chanjo.calculate.cli:calculate',
        ]
    },

    # See: http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are:
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering :: Bio-Informatics',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',

        'Environment :: Console',
    ],
)



# -*- coding: utf-8 -*-
from chanjo.store.models import TranscriptStat
from chanjo.load import sambamba


def test_load_transcripts(exon_lines):
    # GIVEN sambamba depth output lines
    # WHEN loading transcript stats
    result = sambamba.load_transcripts(exon_lines, sample_id='sample',
                                       group_id='group')
    # THEN transcript models should be generated
    assert result.count == 10
    assert result.sample.id == 'sample'
    assert result.sample.group_id == 'group'
    assert isinstance(list(result.models)[0], TranscriptStat)

    # GIVEN no explicit sample id
    # WHEN loading transcript stats
    result = sambamba.load_transcripts(exon_lines)
    # THEN it should be picked up from file
    assert result.sample.id == 'ADM992A10'


def test_load_transcripts_with_threshold(exon_lines):
    # GIVEN a cutoff for "complete" exons at 100x
    threshold = 100
    # WHEN loading transcript stats
    result = sambamba.load_transcripts(exon_lines, threshold=threshold)
    # THEN some transcripts will have incomplete exons linked
    incompletes = [transcript for transcript in result.models
                   if transcript.incomplete_exons]
    assert len(incompletes) > 0



# -*- coding: utf-8 -*-
import sys

import click


def groupby_tx(exons, sambamba=False):
    """Group (unordered) exons per transcript."""
    transcripts = {}
    for exon in exons:
        if sambamba:
            exon['elements'] = dict(zip(exon['extraFields'][-2].split(','),
                                        exon['extraFields'][-1].split(',')))
        else:
            exon['elements'] = dict(exon['elements'])
        for transcript_id in exon['elements']:
            if transcript_id not in transcripts:
                transcripts[transcript_id] = []
            transcripts[transcript_id].append(exon)
    return transcripts


def validate_stdin(context, param, value):
    """Validate piped input contains some data.

    Raises:
        click.BadParameter: if STDIN is empty
    """
    # check if input is a file or stdin
    if value.name == '<stdin>':  # pragma: no cover
        # raise error if stdin is empty
        if sys.stdin.isatty():
            raise click.BadParameter('you need to pipe something to stdin')
    return value



# -*- coding: utf-8 -*-
"""Parse BED files including Sambamba output files."""
from chanjo.compat import zip
from chanjo.exc import BedFormattingError


def chanjo(handle):
    """Parse the chanjo specific columns in a BED file.

    Args:
        handle (iterable): Chanjo-formatted BED lines

    Yields:
        dict: representation of row in BED file
    """
    lines = (line.strip() for line in handle if not line.startswith('#'))
    rows = (line.split('\t') for line in lines)
    for row in rows:
        yield expand_row(row)


def expand_row(row):
    """Parse a BED row.

    Args:
        List[str]: BED row

    Returns:
        dict: formatted BED row as a dict

    Raises:
        BedFormattingError: failure to parse first three columns
    """
    try:
        data = {
            'chrom': row[0],
            'chromStart': int(row[1]),
            'chromEnd': int(row[2])
        }
    except IndexError:
        raise BedFormattingError('make sure fields are tab-separated')
    except ValueError:
        raise BedFormattingError("positions malformatted: {}".format(row))

    try:
        data['score'] = int(list_get(row, 4))
    except ValueError:
        raise BedFormattingError('invalid BED syntax, score column not int')

    data['name'] = list_get(row, 3)
    data['strand'] = list_get(row, 5)
    element_combos = extra_fields(row[6:8])
    data['elements'] = element_combos
    data['extra_fields'] = row[8:]
    return data


def extra_fields(columns):
    """Parse additional, chanjo specific fields.

    Contains transcript and gene ids.

    Args:
        columns (List[str]): list of two columns with related element ids

    Returns:
        List[tuple]: list of tuples with paired elements
    """
    try:
        transcripts = columns[0].split(',')
        genes = columns[1].split(',')
    except IndexError:
        return []
    return zip(transcripts, genes)


def list_get(list_obj, index, default=None):
    """Like ``.get`` for list object.

    Args:
        list_obj (list): list to look up an index in
        index (int): index position to look up
        default (Optional[object]): default return value. Defaults to None.

    Returns:
        object: any object found at the index or ``default``
    """
    try:
        return list_obj[index]
    except IndexError:
        return default
